@@ -1,13 +1,15 @@
 from __future__ import absolute_import
 from __future__ import print_function
 
-import json
 import os
 import sys
 import tempfile
 import urllib
+
+import simplejson
 from PIL import Image
 import numpy as np
+import collections
 
 class JobModel:
     """
@@ -46,6 +48,16 @@ class JobModel:
 
         return None
 
+    @property
+    def resources(self):
+        if 'resources' in self.config:
+            return self.config['resources']
+
+        return {}
+
+    def has_dpu(self):
+        return 'dpu' in self.resources and self.resources['dpu'] > 0
+
     def get_batch_size(self):
         return self.job['config']['batchSize']
 
@@ -196,7 +208,6 @@ class JobModel:
         return image
 
     def get_model_provider(self):
-
         sys.path.append('./')
 
         model = __import__('model', '')
@@ -209,7 +220,7 @@ class JobModel:
     def layers(self):
         if self._layers is None:
             with open(self.get_layers_path(), 'r') as f:
-                self._layers = json.loads(f.read())
+                self._layers = simplejson.loads(f.read(), object_pairs_hook=collections.OrderedDict)
 
         return self._layers
 
@@ -221,7 +232,7 @@ class JobModel:
             def read_dataset(id):
                 dataset_json_path = os.path.normpath(self.get_dataset_dir() + '/' + id + '.json')
                 with open(dataset_json_path, 'r') as f:
-                    self._datasets[id] = json.loads(f.read())
+                    self._datasets[id] = simplejson.loads(f.read(), object_pairs_hook=collections.OrderedDict)
 
             for owner in os.listdir(self.get_dataset_dir()):
                 path = os.path.normpath(self.get_dataset_dir() + '/' + owner + '/dataset/')
@@ -248,7 +259,7 @@ class JobModel:
             if 'datasetId' in layer and layer['datasetId']:
 
                 if layer['datasetId'] not in self.datasets:
-                    raise Exception('Dataset %s not found in datasets %s' % (layer['datasetId'], json.dumps(self.datasets.keys())))
+                    raise Exception('Dataset %s not found in datasets %s' % (layer['datasetId'], simplejson.dumps(self.datasets.keys())))
 
                 dataset = self.datasets[layer['datasetId']]
                 if not dataset:
