@@ -1,29 +1,29 @@
+import collections
 import requests
 import six
 import sys
 
 from requests.auth import HTTPBasicAuth
 from six.moves.urllib.parse import urlencode
-import json
+import simplejson
 
 from aetros.logger import drain_stream
 from aetros.utils import read_home_config, create_ssh_stream
 
-
 class ApiError(Exception):
-    def __init__(self, message, reason):
+    def __init__(self, message, error):
         self.message = message
-        self.reason = reason
+        self.error = error
 
     def __str__(self):
-        return self.message + ', Reason: ' + self.reason
+        return self.message + ', Reason: ' + self.error
 
 
 class ApiConnectionError(ApiError):
     pass
 
 
-def http_request(path, query='', json_body=None, method='get', config=None):
+def http_request(path, query='', json_body=None, method='get', config=None, handle_common_errors=True):
     config = read_home_config() if config is None else config
 
     try:
@@ -40,7 +40,8 @@ def http_request(path, query='', json_body=None, method='get', config=None):
         else:
             path += '?' + query
 
-    url = 'https://' + config['host'] + '/api/' + path
+    url = config['url'] + '/api/' + path
+
     auth = None
     if 'auth_user' in config:
         auth = HTTPBasicAuth(config['auth_user'], config['auth_pw'])
@@ -48,14 +49,23 @@ def http_request(path, query='', json_body=None, method='get', config=None):
     if json_body is not None and method == 'get':
         method = 'post'
 
-    response = requests.request(
-        method, url, data=json_body,
-        auth=auth, verify=config['ssl_verify'],
-        headers={'Accept': 'application/json'}
-    )
+
+    try:
+        response = requests.request(
+            method, url, data=json_body,
+            auth=auth, verify=config['ssl_verify'],
+            headers={'Accept': 'application/json'}
+        )
+    except requests.exceptions.SSLError:
+        if not handle_common_errors:
+            raise
+
+        print("Error: Could not connect to " + url + ". Make sure to install a valid SSL cert or disable ssl check by"
+                                                     "setting aetros home-config ssl_verify false")
+        sys.exit(1)
 
     if response.status_code >= 400:
-        raise_response_exception('Failed request ' + path, response)
+        raise_response_exception('Failed request ' + url, response)
 
     return parse_json(response.content.decode('utf-8'))
 
@@ -77,10 +87,10 @@ def request(path, query=None, body=None, method='get', config=None):
         method = 'post'
 
     ssh_stream = create_ssh_stream(config)
-    stdin, stdout, stderr = ssh_stream.exec_command('api ' + method + ' ' + json.dumps(path))
+    stdin, stdout, stderr = ssh_stream.exec_command('api ' + method + ' ' + simplejson.dumps(path))
 
     if body is not None:
-        input = six.b(json.dumps(body))
+        input = six.b(simplejson.dumps(body))
         stdin.write(input)
         stdin.flush()
         stdin.channel.shutdown_write()
@@ -89,9 +99,10 @@ def request(path, query=None, body=None, method='get', config=None):
     stderr = drain_stream(stderr)
 
     if len(stderr) > 0:
-        stderr = stderr.decode('utf-8')
+        if hasattr(stderr, 'decode'):
+            stderr = stderr.decode('utf-8')
 
-        raise ApiError('Could not request api: ' + path, stderr)
+        raise ApiError('Could not request api: ' + config['host'] + path, stderr)
 
     return stdout
 
@@ -103,7 +114,7 @@ def raise_response_exception(message, response):
 
     if isinstance(content, six.string_types):
         try:
-            content_parsed = json.loads(content)
+            content_parsed = simplejson.loads(content, object_pairs_hook=collections.OrderedDict)
             if 'error' in content_parsed:
                 error = content_parsed['error']
             if 'message' in content_parsed:
@@ -116,10 +127,10 @@ def raise_response_exception(message, response):
 
 
 def parse_json(content):
-    a = json.loads(content)
+    a = simplejson.loads(content)
 
     if isinstance(a, dict) and 'error' in a:
-        raise Exception('API request failed %s: %s.' % (a['error'], a['message']))
+        raise ApiError('API request failed %s: %s.' % (a['error'], a['message']), a['error'])
 
     return a
 
@@ -144,6 +155,6 @@ def create_job(model_name, local=False, parameters=None, dataset_id=None, config
     return parse_json(content)
 
 
-def create_model(model_name, private = False):
-    content = request('model/create', None, {'name': model_name, 'private': private}, 'put')
+def create_model(model_name, organisation=None, space=None, private = False):
+    content = request('model/create', None, {'name': model_name, 'organisation': organisation, 'space': space, 'private': private}, 'put')
     return parse_json(content)
