@@ -1,6 +1,7 @@
 from __future__ import absolute_import
 from __future__ import print_function
 import argparse
+import atexit
 import os
 import sys
 import time
@@ -9,22 +10,19 @@ from threading import Lock
 import paramiko as paramiko
 import psutil
 import subprocess
-
 import signal
-
-import requests
 import six
 from cryptography.hazmat.primitives import serialization
-from requests.auth import HTTPBasicAuth
 
 import aetros.api
-from aetros.git import Git
+from aetros.client import BackendClient
 from aetros.logger import GeneralLogger
 
-from aetros.backend import EventListener, BackendClient
+from aetros.backend import EventListener
 from aetros.utils import unpack_simple_job_id, read_home_config
 import aetros.cuda_gpu
 
+
 class ServerClient(BackendClient):
     def __init__(self, config, event_listener, logger):
         BackendClient.__init__(self, config, event_listener, logger)
@@ -34,9 +32,9 @@ class ServerClient(BackendClient):
     def configure(self, server_name):
         self.server_name = server_name
 
-    def on_connect(self, reconnect=False):
-        self.send_message({'type': 'register_server', 'server': self.server_name, 'reconnect': reconnect})
-        messages = self.wait_for_at_least_one_message()
+    def on_connect(self, reconnect, channel):
+        self.send_message({'type': 'register_server', 'server': self.server_name, 'reconnect': reconnect}, channel)
+        messages = self.wait_for_at_least_one_message(channel)
 
         if not messages:
             return False
@@ -54,7 +52,7 @@ class ServerClient(BackendClient):
                 return False
 
             if "registered" in message['a']:
-                self.registered = True
+                self.registered[channel] = True
                 self.event_listener.fire('registration', {'username': message['username'], 'server': self.server_name})
                 self.handle_messages(messages)
                 return True
@@ -118,7 +116,7 @@ class ServerCommand:
         parser.add_argument('name', nargs='?', help="Server name")
         parser.add_argument('--generate-ssh-key', help="Generates automatically a ssh key, register them in AETROS in "
                                                        "your account, and delete them when the server exits. "
-                                                       "You should prefer 'aetros register' command as its safer.")
+                                                       "You should prefer 'aetros authenticate' command as its safer.")
 
         parser.add_argument('--allow-host-execution', action='store_true', help="Whether a job can run on this server "
             "directly, without a virtual (docker) container.\nSecurity risk and makes resource limitation useless.")
@@ -128,7 +126,7 @@ class ServerCommand:
         parser.add_argument('--max-cpus',
             help="How many cores are available. Per default all available CPU cores.")
         parser.add_argument('--max-gpus',
-            help="How many GPUs are available. Comma separate list of device ids (pciBusId)."
+            help="How many GPUs are available. Comma separate list of device ids."
                  "Per default all available GPU cards. Use 'aetros gpu' too see the ids.")
 
         parser.add_argument('--no-gpus', action='store_true', help="Disable all GPUs")
@@ -161,7 +159,7 @@ class ServerCommand:
             gpus = aetros.cuda_gpu.get_ordered_devices()
             for i in range(len(gpus)):
                 self.enabled_gpus.append(i)
-        except Exception: pass
+        except aetros.cuda_gpu.CudaNotImplementedException: pass
 
         if parsed_args.max_gpus:
             self.enabled_gpus = []
@@ -207,7 +205,14 @@ class ServerCommand:
                 'key': self.ssh_key_public,
             }
 
-            response = aetros.api.http_request('server/ssh-key', json_body=data, method='post')
+            try:
+                response = aetros.api.http_request('server/ssh-key', json_body=data, method='post')
+            except aetros.api.ApiError as e:
+                if 'access_denied' in e.error:
+                    print("error: Could not connect to " + self.config['url'] +
+                          ': Access denied. --generate-ssh-key seems to be wrong. Incorrect host? See "aetros id"')
+                    sys.exit(1)
+                raise
 
             ssh_key_registered = response == True
 
@@ -223,7 +228,6 @@ class ServerCommand:
                 self.logger.error('Could not delete SSH key in AETROS Trainer.')
 
         if parsed_args.generate_ssh_key and ssh_key_registered:
-            import atexit
             atexit.register(delete_ssh_key)
 
         if parsed_args.host:
@@ -248,13 +252,16 @@ class ServerCommand:
         try:
             while self.active:
                 if self.registered:
-                    self.server.send_message({'type': 'utilization', 'values': self.collect_system_utilization()})
+                    self.server.send_message({'type': 'utilization', 'values': self.collect_system_utilization()}, '')
                     self.check_finished_jobs()
 
                 time.sleep(1)
-        except KeyboardInterrupt:
-            self.logger.warning('Aborted')
+        except SystemExit:
+            self.logger.warning('Killed')
             self.stop()
+        except KeyboardInterrupt:
+            self.stop()
+
 
     def on_signusr1(self, signal, frame):
         self.logger.info("ending=%s, active=%s, registered=%s, %d running, %d messages, %d connection_tries" % (
@@ -274,16 +281,19 @@ class ServerCommand:
         self.logger.warning('Closed')
 
     def write_log(self, message):
-        self.server.send_message({'type': 'log', 'message': message})
+        self.server.send_message({'type': 'log', 'message': message}, '')
         return True
 
     def stop(self):
         self.active = False
 
+        self.logger.warning('Killing %d jobs ' % (len(self.job_processes),))
+
         for p in six.itervalues(self.job_processes):
-            p.kill()
-            time.sleep(0.1)
-            p.terminate()
+            p.send_signal(signal.SIGINT)
+
+        for p in six.itervalues(self.job_processes):
+            p.wait()
 
         self.general_logger_stdout.flush()
         self.general_logger_stderr.flush()
@@ -336,7 +346,7 @@ class ServerCommand:
                     reason = 'Failed job %s. Exit status: %s' % (full_job_id, str(exit_code))
                     self.logger.error(reason)
 
-                self.server.send_message({'type': 'job-finished', 'id': full_job_id})
+                self.server.send_message({'type': 'job-finished', 'id': full_job_id}, '')
                 delete_finished.append(full_job_id)
 
         for full_job_id in delete_finished:
@@ -360,9 +370,12 @@ class ServerCommand:
                 for gpu_id in resources_assigned['gpus']:
                     args += ['--gpu-device', gpu_id]
 
+            args += ['--cpu', str(int(resources_assigned['cpus']))]
+            args += ['--memory', str(int(resources_assigned['memory']))]
+
             args += [full_id]
             self.logger.info('$ ' + ' '.join(args))
-            self.server.send_message({'type': 'job-executed', 'id': full_id})
+            self.server.send_message({'type': 'job-executed', 'id': full_id}, '')
 
             # Since JobBackend sends SIGINT to its current process group, wit sends also to its parents when same pg.
             # We need to change the process group of the process, so this won't happen.
@@ -373,8 +386,10 @@ class ServerCommand:
             else:
                 kwargs['preexec_fn'] = os.setsid
 
+            stdout = subprocess.PIPE if self.show_stdout else open(os.devnull, 'r+b', 0)
+
             process = subprocess.Popen(args, bufsize=1, env=my_env, stdin=DEVNULL,
-                stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kwargs)
+                stderr=stdout, stdout=stdout, **kwargs)
 
             if self.show_stdout:
                 self.general_logger_stdout.attach(process.stdout, read_line=True)
@@ -385,7 +400,7 @@ class ServerCommand:
     def registration_complete(self, params):
         self.registered = True
         self.logger.info("Server connected to %s as %s under account %s registered." % (self.config['host'], params['server'], params['username']))
-        self.server.send_message({'type': 'system', 'values': self.collect_system_information()})
+        self.server.send_message({'type': 'system', 'values': self.collect_system_information()}, '')
 
     def collect_system_information(self):
         values = {}
@@ -407,7 +422,7 @@ class ServerCommand:
                 gpu['available'] = gpu_id in self.enabled_gpus
 
                 values['gpus'][gpu_id] = gpu
-        except Exception: pass
+        except aetros.cuda_gpu.CudaNotImplementedException: pass
 
         for disk in psutil.disk_partitions():
             try:
@@ -450,7 +465,7 @@ class ServerCommand:
         try:
             for gpu_id, gpu in enumerate(aetros.cuda_gpu.get_ordered_devices()):
                 values['gpus'][gpu_id] = aetros.cuda_gpu.get_memory(gpu['device'])
-        except Exception: pass
+        except aetros.cuda_gpu.CudaNotImplementedException: pass
 
         for disk in psutil.disk_partitions():
             try:
@@ -460,6 +475,8 @@ class ServerCommand:
 
         net_stats = psutil.net_io_counters(pernic=True)
         for id in self.nets:
+            if id not in net_stats:
+                continue
             net = net_stats[id]
             values['nets'][id] = {
                 'recv': net.bytes_recv,
