@@ -1,10 +1,13 @@
 from __future__ import division
 from __future__ import absolute_import
 
+import logging
 import os
 import re
 import time
 import datetime
+import traceback
+from threading import Thread
 
 import numpy as np
 import signal
@@ -13,10 +16,14 @@ import paramiko
 import six
 import sys
 import ruamel.yaml as yaml
+import subprocess
+
+from paramiko.compress import ZlibCompressor
 
 start_time = time.time()
 last_time = None
 
+
 def stop_time(title=''):
     global last_time
     diff = ("{0:.10f}".format(time.time() - last_time)) if last_time is not None else ''
@@ -24,6 +31,89 @@ def stop_time(title=''):
     sys.__stdout__.write("STOP_TIME: " + str(time.time()-start_time) + "s - diff: "+diff+"  - " +title+ "\n")
 
 
+def is_debug2():
+    return os.getenv('DEBUG') == '2'
+
+
+def is_debug():
+    return is_debug2() or os.getenv('DEBUG') == '1'
+
+
+def get_logger(name='', debug=None, format=None):
+
+    import coloredlogs
+    # logging.basicConfig() # this will make paramiko log a lot of stuff
+    logger = logging.getLogger(name if name else 'aetros')
+
+    level = 'INFO'
+    fmt = '%(message)s' if format is None else format
+
+    if debug is None:
+        debug = is_debug()
+
+    if debug:
+        level = 'DEBUG'
+        if format is None:
+            fmt = coloredlogs.DEFAULT_LOG_FORMAT
+
+    atty = None
+    if '1' == os.getenv('AETROS_ATTY'):
+        atty = True
+
+    coloredlogs.install(fmt=fmt, level=level, logger=logger, isatty=atty)
+
+    return logger
+
+
+def ensure_docker_installed(logger, home_config=None):
+    home_config = home_config or read_home_config()
+    try:
+        out = subprocess.check_output([home_config['docker'], '-v'])
+        return True
+    except Exception as e:
+        logger.error('Docker is not installed: ' + (str(e)))
+        sys.exit(2)
+
+
+def docker_call(args, home_config=None):
+    home_config = home_config or read_home_config()
+    base = [home_config['docker']]
+    return subprocess.check_output(base + args)
+
+
+def loading_text(label='Loading ... '):
+    import itertools, sys
+    spinner = itertools.cycle(['-', '/', '|', '\\'])
+    state = {'active': True}
+
+    sys.stdout.write(label)
+    sys.stdout.flush()
+
+    def display_thread(state):
+        try:
+            while state['active']:
+                sys.stdout.write(next(spinner))
+                sys.stdout.flush()
+                time.sleep(0.05)
+                sys.stdout.write('\b')
+        except (KeyboardInterrupt, SystemExit):
+            return
+
+    thread = None
+    if not is_debug():
+        thread = Thread(target=display_thread, args=[state])
+        thread.daemon = True
+        thread.start()
+
+    def stop(done_label="done."):
+        state['active'] = False
+        thread and thread.join()
+        sys.stdout.write(done_label + '\n')
+        sys.stdout.flush()
+
+    return stop
+
+
 def get_option(dict, key, default=None, type=None):
     if key not in dict or dict[key] == '':
         return default
@@ -34,6 +124,60 @@ def get_option(dict, key, default=None, type=None):
     return dict[key]
 
 
+def extract_api_calls(line, callback, print_traceback=False, logger=None):
+    failed_calls = []
+    handled_calls = []
+    filtered_line = line
+
+    if hasattr(filtered_line, 'decode'):
+        filtered_line = filtered_line.decode('utf-8')
+
+    c = 0
+    while True:
+        start_pos = filtered_line.find('{aetros:')
+        if -1 == start_pos:
+            start_pos = filtered_line.find('{"aetros":')
+
+        end_pos = filtered_line.find('}\n')
+        eat_end = 2
+        if -1 == end_pos:
+            end_pos = filtered_line.find('}\r\n')
+            eat_end = 3
+            if -1 == end_pos:
+                end_pos = filtered_line.find('}\r')
+                eat_end = 2
+
+        if start_pos == -1 or end_pos == -1:
+            break
+
+        line = filtered_line[start_pos:end_pos+1]
+
+        try:
+            call = yaml.load(line, Loader=yaml.RoundTripLoader)
+            if callback(call) is False:
+                failed_calls.append({'line': line, 'exception': Exception('Unknown API call.')})
+            else:
+                handled_calls.append(call)
+        except (KeyboardInterrupt, SystemExit):
+            raise
+        except Exception as e:
+            if print_traceback:
+                sys.__stderr__.write(traceback.format_exc())
+            else:
+                logger and logger.debug(traceback.format_exc())
+            failed_calls.append({'line': line, 'exception': e})
+
+        filtered_line = filtered_line[0:start_pos] + filtered_line[end_pos+eat_end:]
+
+        c += 1
+
+        if c > 10:
+            # allow max 10 in one line
+            break
+
+    return handled_calls, filtered_line, failed_calls
+
+
 def unpack_simple_job_id(full_id):
     [owner, model, id] = unpack_full_job_id(full_id)
 
@@ -51,6 +195,10 @@ def unpack_full_job_id(full_id):
     return [owner, model, id]
 
 
+class KeyNotConfiguredException(Exception):
+    pass
+
+
 def create_ssh_stream(config, exit_on_failure=True):
     ssh_stream = paramiko.client.SSHClient()
     # ssh_stream.load_system_host_keys()
@@ -64,18 +212,30 @@ def create_ssh_stream(config, exit_on_failure=True):
         key = paramiko.RSAKey.from_private_key(six.StringIO(config['ssh_key_base64']))
 
     if not key and not key_filename:
-        raise Exception("No SSH key configured for " + config['host'] + ". See https://aetros.com/docu/trainer/authentication")
+        raise KeyNotConfiguredException("No SSH key configured for " + config['host']
+                                        + ". See https://aetros.com/docu/trainer/authentication")
 
     key_description = key_filename if key_filename else 'from server'
 
     try:
-        ssh_stream.connect(config['host'], key_filename=key_filename, username='git', compress=True, pkey=key)
+        def lower_compression(self):
+            import zlib
+            # Use the default level of zlib compression
+            self.z = zlib.compressobj(2)
+
+        ZlibCompressor.__init__ = lower_compression
+
+        ssh_stream.connect(config['host'], port=config['ssh_port'], key_filename=key_filename, username='git', compress=True, pkey=key)
+        # ssh_stream.get_transport().window_size = 2147483647
+    except (SystemExit, KeyboardInterrupt):
+        raise
     except Exception as e:
-        if isinstance(e, paramiko.ssh_exception.AuthenticationException) or isinstance(e, paramiko.ssh_exception.SSHException):
-            if exit_on_failure:
-                print("Fatal: AETROS authentication against "+config['host']+" failed using key "+key_description+": "+str(e)+
-                      ". Did you setup SSH keys correctly? See https://aetros.com/docu/trainer/authentication")
-                sys.exit(1)
+        if exit_on_failure:
+            message = str(type(e).__name__) + ': ' + str(e)
+            sys.stdout.write("Fatal: AETROS authentication against "+config['host']+" failed using key " +
+                             key_description + ": " + message +
+                             ". Did you setup SSH keys correctly? See https://aetros.com/docu/trainer/authentication\n")
+            sys.exit(1)
         raise
 
     return ssh_stream
@@ -84,6 +244,7 @@ def create_ssh_stream(config, exit_on_failure=True):
 def setup_git_ssh(config):
     import tempfile
     ssh_command = config['ssh']
+    ssh_command += ' -p ' + str(config['ssh_port'])
     ssh_command += ' -o StrictHostKeyChecking=no'
 
     ssh_key = None
@@ -143,27 +304,54 @@ def read_home_config(path = None, logger=None):
             logger and logger.debug('Home config loaded from ' + os.path.realpath(path))
             custom_config = yaml.safe_load(f)
         except Exception:
-            raise Exception('Could not load aetros home config at ' + os.path.realpath(path))
+            sys.stderr.write('Error: could not load aetros home config at ' + os.path.realpath(path) + '\n')
+            raise
 
         if custom_config is None:
             custom_config = {}
 
     config = {
-        'host': os.getenv('AETROS_HOST') or 'trainer.aetros.com',
-        'storage_dir': os.getenv('AETROS_STORAGE_DIR') or '~/.aetros',
-        'ssh_key_base64': os.getenv('AETROS_SSH_KEY_BASE64'),
+        'host': 'trainer.aetros.com',
+        'storage_dir': '~/.aetros',
+        'ssh_key_base64': None,
         'image': None,
         'ssh': 'ssh',
         'git': 'git',
         'docker': 'docker',
         'docker_options': [],
+        'http_port': 80,
+        'https_port': 443,
+        'ssh_port': 22,
+        'ssl': True,
         'ssl_verify': True,
     }
 
     config.update(custom_config)
 
+    if os.getenv('AETROS_HOST'):
+        config['host'] = os.getenv('AETROS_HOST')
+
+    if os.getenv('AETROS_STORAGE_DIR'):
+        config['storage_dir'] = os.getenv('AETROS_STORAGE_DIR')
+
+    if os.getenv('AETROS_SSH_KEY_BASE64'):
+        config['ssh_key_base64'] = os.getenv('AETROS_SSH_KEY_BASE64')
+
     config['storage_dir'] = os.path.abspath(os.path.expanduser(config['storage_dir']))
 
+    http = 'https://'
+    host = config['host']
+
+    if config['https_port'] != 443:
+        host += ':' + str(config['https_port'])
+
+    if not config['ssl']:
+        http = 'http://'
+        if config['http_port'] != 80:
+            host += ':' + str(config['http_port'])
+
+    config['url'] = http + host
+
     return config
 
 ignore_pattern_cache = {}
@@ -177,6 +365,7 @@ def extract_parameters(full_definitions, overwritten = None, incoming_path = '')
 
     for parameter in full_definitions:
         param_type = parameter['type']
+        param_subtype = parameter['subtype'] if 'subtype' in parameter else None
 
         if 'defaultValue' in parameter:
             defaultValue = parameter['defaultValue']
@@ -200,7 +389,10 @@ def extract_parameters(full_definitions, overwritten = None, incoming_path = '')
             container[name] = str(value)
 #
         if param_type == 'number':
-            container[name] = float(value)
+            if param_subtype == 'int':
+                container[name] = int(value)
+            else:
+                container[name] = float(value)
 #
         if param_type == 'boolean':
             container[name] = bool(value)
@@ -281,6 +473,7 @@ def lose_parameters_to_full(parameters):
                 if not v:
                     definition['type'] = 'choice_string'
                     continue
+
                 if isinstance(v[0], six.string_types):
                     definition['type'] = 'choice_string'
                 else:
@@ -313,6 +506,9 @@ def lose_parameters_to_full(parameters):
                 definition['children'] = extract_from_parameters(v)
             else:
                 definition['type'] = type_of(v)
+                if isinstance(v, (int)):
+                    definition['subtype'] = 'int'
+
                 definition['defaultValue'] = v
 
             definitions.append(definition)
@@ -408,9 +604,9 @@ def is_ignored(path, ignore_patters):
             regex = regex.replace('\\*', '[^/\\\\]+')
             regex = '(' + regex + ')'
             if pattern[0] == '/' or (pattern[0] == '!' and pattern[1] == '/'):
-                regex = '^' + regex
+                regex = '^' + regex + '$'
             else:
-                regex = '^.*' + regex
+                regex = '^.*' + regex + '$'
 
             reobj = re.compile(regex)
             ignore_pattern_cache[pattern] = reobj
@@ -437,10 +633,111 @@ def is_ignored(path, ignore_patters):
     return ignored
 
 
-def read_config(path = 'aetros.yml', logger=None):
-    path = os.path.normpath(os.path.expanduser(path))
+def git_local_job_ids(home_config, model):
+    git_dir = os.path.normpath(home_config['storage_dir'] + '/' + model + '.git')
 
-    config = {
+    output = subprocess.check_output([home_config['git'], '--bare', '--git-dir', git_dir, 'show-ref']).decode('utf-8')
+
+    job_ids = []
+
+    for line in output.split('\n'):
+        if not ' ' in line:
+            continue
+        target, ref_name = line.split(' ')
+        if ref_name.startswith('refs/aetros/job/'):
+            job_ids.append(ref_name[len('refs/aetros/job/'):])
+
+    return job_ids
+
+
+def git_remote_job_ids(home_config, model):
+    git_remote_url = 'git@%s:%s.git' % (home_config['host'], model)
+    output = subprocess.check_output([home_config['git'], 'ls-remote', git_remote_url]).decode('utf-8')
+    job_ids = []
+
+    for line in output.split('\n'):
+        if '\t' not in line:
+            continue
+        target, ref_name = line.split('\t')
+        if ref_name.startswith('refs/aetros/job/'):
+            job_ids.append(ref_name[len('refs/aetros/job/'):])
+
+    return job_ids
+
+
+def git_has_local_job(home_config, model, job_id):
+    job_ids = git_local_job_ids(home_config, model)
+    for full_job_id in job_ids:
+        if full_job_id.startswith(job_id):
+            return full_job_id
+
+
+def git_has_remote_job(home_config, model, job_id):
+    job_ids = git_remote_job_ids(home_config, model)
+    for full_job_id in job_ids:
+        if full_job_id.startswith(job_id):
+            return full_job_id
+
+
+def find_config(path = None, error_on_missing=False, return_default=True, logger=None):
+    config = None
+
+    if path:
+        if os.path.exists(path):
+            config = read_config(path, return_default=False, logger=logger)
+            config['init_config_path'] = os.path.abspath(path)
+
+    else:
+        path = find_config_path()
+        if path:
+            config = read_config(path, return_default=False, logger=logger)
+            config['init_config_path'] = os.path.abspath(path)
+
+    if config:
+        if 'import' in config and config['import']:
+            inherited_config = find_config(os.path.dirname(os.path.abspath(path)) + '/' + config['import'])
+
+            if inherited_config:
+                inherited_config.update(config)
+                config = inherited_config
+
+        if config['model']:
+            if 'init_config_path' in config:
+                config['working_dir'] = os.path.relpath(os.path.dirname(config['init_config_path']), config['root'])
+                config['configPath'] = os.path.relpath(config['init_config_path'], config['root'])
+                del config['init_config_path']
+
+            if return_default:
+                return apply_config_defaults(config)
+            else:
+                return config
+
+    if error_on_missing:
+        sys.stderr.write('Error: No AETROS Trainer model name given. Specify it in aetros.yml '
+                         '"model: user/model-name" or use "aetros init model-name".\n')
+        sys.exit(2)
+
+    return {'model': None}
+
+
+def find_config_path(dir = None):
+    if not dir:
+        dir = os.path.abspath(os.getcwd())
+
+    while True:
+        if os.path.exists(dir+ '/aetros.yml'):
+            return dir + '/aetros.yml'
+        else:
+            new_path = os.path.realpath(dir + '/..')
+            if new_path == dir:
+                return None
+
+            dir = new_path
+
+
+def apply_config_defaults(config):
+    defaults = {
+        'model': None,
         'dockerfile': None,
         'command': None,
         'install': None,
@@ -448,26 +745,41 @@ def read_config(path = 'aetros.yml', logger=None):
         'image': None,
         'server': None,
         'parameters': {},
+        'import': None,
+        'root': os.getcwd(),
+        'working_dir': None,
         'servers': None,
+        'configPath': None,
         'before_command': [],
     }
 
+    defaults.update(config)
+
+    return defaults
+
+
+def read_config(path='aetros.yml', logger=None, return_default=True):
+    path = os.path.normpath(os.path.abspath(os.path.expanduser(path)))
+
+    config = {}
+
     if os.path.exists(path):
         f = open(path, 'r')
 
-        custom_config = yaml.safe_load(f)
-        if custom_config is None:
-            custom_config = {}
+        config = yaml.load(f, Loader=yaml.RoundTripLoader)
+        if config is None:
+            config = {}
 
-        if 'storage_dir' in custom_config:
-            del custom_config['storage_dir']
+        if 'storage_dir' in config:
+            del config['storage_dir']
 
-        config.update(custom_config)
+        if 'model' in config and config['model']:
+            config['root'] = os.path.dirname(path)
 
         logger and logger.debug('Config loaded from ' + os.path.realpath(path))
 
-    if 'parameters' not in config:
-        config['parameters'] = {}
+    if return_default:
+        return apply_config_defaults(config)
 
     return config
 
